# UpdateManager.py
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-
import sys
import time
import dbus
import logging
import dbus.service
import traceback
from gettext import gettext as _
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

from .Core.Database import Sqlite3Server
from .Core.loop import mainloop
from SystemUpdater.Core.UpdaterConfigParser import UpgradeConfig
from SystemUpdater.packageManager.UpdaterSDKInterface import UpdaterSDKPkgInstall
from .UpgradeStrategiesDbus import UpgradeStrategiesDbusController, \
                                UPDATER_DBUS_INTERFACE, \
                                UPDATER_DBUS_PATH, \
                                UPDATER_DBUS_SERVICE
from SystemUpdater.packageManager.pmUtils import *

STRATEGY_IDLE_INTERVAL = 2*60
STRATEGY_IDLE_TIMEOUT = 6*60
class UpgradeStrategies():
    def __init__(self,options):
        try:
            self.options            = options
            self.bus                = dbus.SystemBus ()
            self.dbus_send          = self._setup_dbus ( self.bus )
            self.uuconfigs          = UpgradeConfig (
                                        datadir = "/var/lib/unattended-upgrades/",
                                        name = "unattended-upgrades-policy.conf" )
            self.configs_uncover    = UpgradeConfig (
                                        defaults_conf = "/etc/kylin-system-updater/system-updater-defaults.conf" )
            self.sqlite3_server     = Sqlite3Server ( self, _no_DataMigration=True )

            #策略配置接口的超时退出机制
            self.strategy_timestamp = 0
            # GLib.timeout_add_seconds(STRATEGY_IDLE_INTERVAL,
            #     self._check_strategy_inactivity)

            # self.updater_kconf2 = UpdaterSDKConf2(self)
            self.updater_package = UpdaterSDKPkgInstall ( self )

        except Exception as e:
            logging.error(e)
            traceback.print_exc()
    
    def run(self):
        """Start the daemon and listen for calls."""
        logging.info("Waiting for calls...")
        try:
            mainloop.run()
        except KeyboardInterrupt:
            self.dbus_send.Quit(None)

    def _setup_dbus(self, bus):
        # check if there is another g-a-i already and if not setup one
        # listening on dbus
        try:
            bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                            bus,
                                            do_not_queue=True)
            logging.info("Initiate dbus success ...")
            return  UpgradeStrategiesDbusController(self, bus_name)
        except dbus.exceptions.NameExistsException:
            if self.options.replace is False:
                logging.critical("Another daemon is already running")
                sys.exit(1)
            logging.warning("Replacing already running daemon")
            
            retry_reboot_times = 0
            the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                            UPDATER_DBUS_PATH)
            the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                timeout=300)
            time.sleep(1)
            while True:
                retry_reboot_times = retry_reboot_times + 1
                #当重试次数超过5次时退出程序
                if retry_reboot_times > 5:
                    logging.critical("Reboot backend is Failed...")
                    sys.exit(1)
                try:
                    bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                                    bus,
                                                    do_not_queue=True)
                    logging.warning("Replacing already running daemon to Success...")
                    return  UpgradeStrategiesDbusController(self, bus_name)
                except dbus.exceptions.NameExistsException:
                    the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                                    UPDATER_DBUS_PATH)
                    the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                        timeout=300)
                    logging.error("Dbus has not withdrawn and retry reboot times:%d...",retry_reboot_times)
                    time.sleep(1)
    
    def _check_strategy_inactivity(self):
        logging.info("Checking for inactivity in Strategies daemon ...")
        timestamp = self.strategy_timestamp
        if timestamp == 0:
            self.strategy_timestamp = time.time()
            return True
        #超时退出
        if self.strategy_timestamp != 0 and  time.time() - self.strategy_timestamp > STRATEGY_IDLE_TIMEOUT:
            logging.warning("Quitting due to inactivity")
            self.dbus_send.Quit(None)
            return False
        return True

    def start_deb_install(self, deb_uri = "", _check_local_dep = False, _auto_satisfy = False, source = '', sender=None):
        _success = False
        header = ''
        desc = ''
        sender_name = get_proc_from_dbus_name(sender)
        caller_trans = get_source_name_from_enum(sender_name)

        try:
            def _run_install_handler():
                _success = self.updater_package.package_interface_install(deb_uri)
                if self.updater_package.installFinished:
                    self.updater_package.installFinished = False
                else:
                    self.dbus_send.InstalldebFinished(_success, header, desc)
                    self.dbus_send.InstalldebFinishedWithErrCode(_success, header, desc, '')

            def on_authorization_success(result):
                logging.info("Authorization success ...")

                self.sqlite3_server.deb_policy_keep = True
                def _check_deb_policy():
                    if self.sqlite3_server.deb_policy_timestamp % 10 == 0:
                        logging.info("Checking for deb policy timeout(%d)...",self.sqlite3_server.deb_policy_timestamp)
                    if (self.sqlite3_server.deb_policy_timestamp <= 0):
                        logging.warning("Deb policy timeout")
                        self.sqlite3_server.deb_policy_keep = False
                        return False
                    else:
                        self.sqlite3_server.deb_policy_timestamp = self.sqlite3_server.deb_policy_timestamp - 1
                    return True

                from gi.repository import GLib
                self.sqlite3_server.deb_policy_timestamp = 60 * 5
                GLib.timeout_add_seconds(1,_check_deb_policy)

                _run_install_handler()

            def on_authorization_failure(error):
                try:
                    logging.info("Authorization failed ...")

                    self.sqlite3_server.deb_policy_keep = False
                    self.sqlite3_server.deb_policy_timestamp = 0

                    error.raise_exception()
                except Exception as error:
                    logging.error(error)
                    self.dbus_send.InstalldebFinished(False,_("Cancel authentication."),'')
                    self.dbus_send.InstalldebFinishedWithErrCode(False,_("Cancel authentication."),'','')

            def _policykit_authority_handler(install_handler = None, authority_failed_handler = None):
                if not self.sqlite3_server.deb_policy_keep:
                    deferred = PolicyKit_Authority(caller_trans+_(" requires authentication to install software packages."),
                                                sender = sender,
                                                authentication = self.configs_uncover.getWithDefault(
                                                                                        "InstallAndPurge",
                                                                                        "install_authority",True),
                                                source=source)

                    deferred.add_callback(on_authorization_success)
                    deferred.add_errback(on_authorization_failure)
                else:
                    self.sqlite3_server.deb_policy_timestamp = 60 * 5
                    logging.info("Deb policy keep, ignore...")

                    _run_install_handler()

            if not deb_verify(deb_uri):
                _policykit_authority_handler()
            else:
                _run_install_handler()

        except Exception as e:
            logging.info("start_deb_install error: %s", e)
            self.dbus_send.InstalldebFinished(False, str(e), desc)
            self.dbus_send.InstalldebFinishedWithErrCode(False, str(e), desc, '')

    def start_purge_pkgs(self,pkgs_list,sender=None):
        _success = False
        header = ''
        desc = ''

        try:
            sender_name = get_proc_from_dbus_name(sender)
            caller_trans = get_source_name_from_enum(sender_name)

            def _run_purge_handler():
                for pl in pkgs_list:
                    _success = self.updater_package.package_interface_remove(pl)
                    if not _success:
                        break

                if self.updater_package.purgeFinished:
                    self.updater_package.purgeFinished = False
                else:
                    self.dbus_send.PurgePackagesFinished (_success, header, desc)
                    self.dbus_send.PurgePackagesFinishedWithPkgname (_success, header, desc, ",".join(pkgs_list))
                    self.dbus_send.PurgePackagesFinishedWithErrCode (_success, header, desc, ",".join(pkgs_list),
                                                                    get_error_num_from_enum (SUCCESS_PURGE_PKG))

            def on_authorization_success(result):
                logging.info("Authorization success ...")

                self.sqlite3_server.purge_policy_keep = True
                def _check_purge_policy():
                    if self.sqlite3_server.purge_policy_timestamp % 10 == 0:
                        logging.info("Checking for deb policy timeout(%d)...",self.sqlite3_server.purge_policy_timestamp)
                    if (self.sqlite3_server.purge_policy_timestamp <= 0):
                        logging.warning("Deb policy timeout")
                        self.sqlite3_server.purge_policy_keep = False
                        return False
                    else:
                        self.sqlite3_server.purge_policy_timestamp = self.sqlite3_server.purge_policy_timestamp - 1
                    return True

                from gi.repository import GLib
                self.sqlite3_server.purge_policy_timestamp = 60 * 5
                GLib.timeout_add_seconds(1,_check_purge_policy)

                _run_purge_handler()

            def on_authorization_failure(error):
                """Convert the DeferredException to a normal exception."""
                try:
                    logging.info("Authorization failed ...")

                    self.sqlite3_server.purge_policy_keep = False
                    self.sqlite3_server.purge_policy_timestamp = 0

                    error.raise_exception()
                except Exception as error:
                    logging.error(error)
                    self.dbus_send.PurgePackagesFinished(False,_("Cancel authentication."),'')
                    self.dbus_send.PurgePackagesFinishedWithPkgname(False,_("Cancel authentication."),'',",".join(pkgs_list))
                    self.dbus_send.PurgePackagesFinishedWithErrCode(False,_("Cancel authentication."),'',",".join(pkgs_list),get_error_num_from_enum(ERROR_PURGE_PKG_VERIFY))

            def _policykit_authority_handler(install_handler = None, authority_failed_handler = None):
                if not self.sqlite3_server.purge_policy_keep:
                    deferred = PolicyKit_Authority(
                                    get_source_name_from_enum(sender_name)+_(" requires authentication to uninstall software packages."),
                                    sender = sender,
                                    authentication = self.configs_uncover.getWithDefault("InstallAndPurge","purge_authority",False), 
                                    source=sender_name)

                    deferred.add_callback(on_authorization_success)
                    deferred.add_errback(on_authorization_failure)
                else:
                    self.sqlite3_server.purge_policy_timestamp = 60 * 5
                    logging.info("Purge policy keep, ignore...")

                    _run_purge_handler()

            _policykit_authority_handler()

        except Exception as e:
            logging.error("start_purge_pkgs error: %s", e)

    def start_back_upgrade(self, pkgs_list, sender=None):
        _success        = False
        error_string    = ''
        error_desc      = ''

        try:
            self.updater_package.batchInstPkgs = [str(item) for item in pkgs_list]

            _success = self.updater_package.package_interface_batch_install(pkgs_list)
            if self.updater_package.batchInstFinished:
                self.updater_package.batchInstFinished = False
            else:
                self.dbus_send.UpdateInstallFinished(_success, self.updater_package.batchInstPkgs, error_string, error_desc)
                self.updater_package.package_interface_install_clear()

        except Exception as e:
            logging.error("start_back_upgrade err: %s", e)

    def check_package_status(self, _name, _version = '', sender = None):
        install_status = -100

        try:
            if not _version:
                _version = 'NULL'

            install_status = self.updater_package.package_interface_check_status(_name, _version)

        except Exception as e:
            logging.error("check_package_status err: %s", e)

        return install_status
